
from copy import copy
import sqlite3
import pytest
import pandas_to_sql
from pandas_to_sql.testing.utils import fake_data_creation

sql_connection = sqlite3.connect('./example.db') #create db

TABLE_NAME_1 = 'random_data_1'
DF1, SCHEMA_1 = fake_data_creation.create_fake_dataset()
DF1.to_sql(TABLE_NAME_1, sql_connection, if_exists='replace', index=False, dtype=SCHEMA_1)

TABLE_NAME_2 = 'random_data_2'
DF2, SCHEMA_2 = fake_data_creation.create_fake_dataset()
DF2.columns = DF2.columns.map(lambda c: c + '_2')
DF2.to_sql(TABLE_NAME_2, sql_connection, if_exists='replace', index=False, dtype=SCHEMA_2)

def pytest_configure():
    pytest.df1 = None
    pytest.df2 = None
    pytest.sql_connection = sql_connection

@pytest.fixture(scope="function", autouse=True)
def run_around_tests():
    # print('\nhere\n')
    pytest.df1 = pandas_to_sql.wrap_df(copy(DF1), TABLE_NAME_1)
    pytest.df2 = pandas_to_sql.wrap_df(copy(DF2), TABLE_NAME_2)
    yield
    # run after function



from pandas_to_sql.table import create_table_from_schema
from pandas_to_sql.utils.helpers import  create_schema_from_df
from pandas_to_sql.utils.pandas_dataframe_intercepter import PandasDataFrameIntercepter
from pandas_to_sql.utils.pandas_interceptor import PandasIntercepter


def wrap_df(df, table_name):
    t = create_table_from_schema(table_name=table_name, schema=create_schema_from_df(df))
    return PandasDataFrameIntercepter(df, t)


def wrap_pd(pd):
    return PandasIntercepter(pd)


import numbers
import operator

NUMERIC_COLUMNS_TYPES = ['INT','FLOAT']

def get_type(v):
    if isinstance(v, Column): return v.dtype
    elif isinstance(v, str): return 'VARCHAR'
    elif isinstance(v, int): return 'INT'
    elif isinstance(v, numbers.Number): return 'FLOAT'
    else: raise Exception(f"Value not supported. supporting: premitives and {str(Column)}. got {str(type(v))}")

def is_numeric_type(v_type):
    return v_type in NUMERIC_COLUMNS_TYPES

def validate_numeric(l, r):
    l_type = get_type(l)
    r_type = get_type(r)
    if not is_numeric_type(l_type) or not is_numeric_type(r_type):
        raise Exception(f"supporting only numerics")

def numeric_op_result_from_types(l, r):
    validate_numeric(l, r)
    l_type = get_type(l)
    r_type = get_type(r)
    return 'INT' if l_type is 'INT' and r_type is 'INT' else 'FLOAT'

def value_to_sql_string(value):
    if isinstance(value, numbers.Number):
        return str(value)
    elif isinstance(value, str):
        return "'" + value + "'"
    elif isinstance(value, Column):
        return value.sql_string
    raise Exception(f"Value not supported. supporting: premitives and {str(Column)}. got {str(type(value))}")

def create_column_from_operation(l, r, dtype, op):
    return Column(dtype=dtype,
                    sql_string=f'({value_to_sql_string(l)} {op} {value_to_sql_string(r)})')  

def get_add_op_res_type(l, r):
    l_type = get_type(l)
    r_type = get_type(r)
    if l_type == 'INT' and r_type == 'INT': return 'INT'
    elif l_type == 'VARCHAR' and r_type == 'VARCHAR': return 'VARCHAR'
    elif is_numeric_type(l_type) and is_numeric_type(r_type): return 'FLOAT'
    else: raise Exception(f"add not supprting received types. received: left type: {l_type}, right type: {r_type}")

def get_mul_op_res_type(l, r):
    l_type = get_type(l)
    r_type = get_type(r)
    if l_type == 'INT' and r_type == 'INT': return 'INT'
    elif is_numeric_type(l_type) and is_numeric_type(r_type): return 'FLOAT'
    else: raise Exception(f"mul not supprting received types. received: left type: {l_type}, right type: {r_type}")


class Column:
    dtype = None
    sql_string = None
    
    def __init__(self, dtype, sql_string):
        self.dtype = dtype
        self.sql_string = sql_string
    
    def __copy__(self):
        return Column(self.dtype, self.sql_string)

    def __add__(self, r):
        result_column_type = get_add_op_res_type(self, r)
        op = '||' if result_column_type == 'VARCHAR' else '+'
        return create_column_from_operation(self, r, result_column_type, op)

    def __radd__(self, l):
        result_column_type = get_add_op_res_type(l, self)
        op = '||' if result_column_type == 'VARCHAR' else '+'
        return create_column_from_operation(l, self, result_column_type, op)

    def __sub__(self, r):
        result_column_type = numeric_op_result_from_types(self, r)
        return create_column_from_operation(self, r, result_column_type, '-')

    def __rsub__(self, l):
        result_column_type = numeric_op_result_from_types(l, self)
        return create_column_from_operation(l, self, result_column_type, '-')

    def __mul__(self, r):
        result_column_type = get_mul_op_res_type(self, r)
        return create_column_from_operation(self, r, result_column_type, '*')
    
    def __rmul__(self, l):
        result_column_type = get_mul_op_res_type(l, self)
        return create_column_from_operation(l, self, result_column_type, '*')

    def __truediv__(self, r):
        validate_numeric(self, r)
        return Column(dtype='FLOAT',
                sql_string=f'(({value_to_sql_string(self)} + 0.0) / {value_to_sql_string(r)})')  
    
    def __rtruediv__(self, l):
        validate_numeric(l, self)
        return Column(dtype='FLOAT',
                sql_string=f'(({value_to_sql_string(l)} + 0.0) / {value_to_sql_string(self)})')  

    def __floordiv__(self, r):
        validate_numeric(self, r)
        # http://sqlite.1065341.n5.nabble.com/floor-help-td46158.html
        return Column(dtype='FLOAT',
                sql_string=f'( ROUND(({value_to_sql_string(self)} / {value_to_sql_string(r)}) - 0.5) )')

    def __rfloordiv__(self, l):
        validate_numeric(l, self)
        # http://sqlite.1065341.n5.nabble.com/floor-help-td46158.html
        return Column(dtype='FLOAT',
                sql_string=f'( ROUND(({value_to_sql_string(l)} / {value_to_sql_string(self)}) - 0.5) )')

    def __lt__(self,other):
        return create_column_from_operation(self, other, 'BOOL', '<')

    def __le__(self,other):
        return create_column_from_operation(self, other, 'BOOL', '<=')

    def __gt__(self,other):
        return create_column_from_operation(self, other, 'BOOL', '>')

    def __ge__(self,other):
        return create_column_from_operation(self, other, 'BOOL', '>=')

    def __eq__(self,other):
        return create_column_from_operation(self, other, 'BOOL', '=')

    def __ne__(self,other):
        return create_column_from_operation(self, other, 'BOOL', '<>')
    
    def __invert__(self):
        if self.dtype != 'BOOL':
            raise Exception(f'tilde support only bool columns, got: {self.dtype}')
        return Column(dtype='BOOL',
                sql_string=f'(NOT({value_to_sql_string(self)}))')

    def __abs__(self):
        t = get_type(self)
        if not is_numeric_type(t):
            raise Exception(f'abs support only numeric columns, got: {str(t)}')
        return Column(dtype=t,
                    sql_string=f'ABS({value_to_sql_string(self)})')  

    def __neg__(self):
        t = get_type(self)
        sql_string = None
        if is_numeric_type(t): sql_string = f'(-({value_to_sql_string(self)}))'
        elif t == 'BOOL': sql_string = f'(NOT({value_to_sql_string(self)}))'
        else: raise Exception(f'- support only numeric and bool columns, got: {str(t)}')
        return Column(dtype=t,
                    sql_string=sql_string)  
       






from pandas_to_sql.column import Column
from pandas_to_sql.grouped_table import GroupedTable
from copy import copy

class Table:
    table_name = None
    columns = None
    filters = None
    from_sql_string = None
    had_changed = None
    
    def __init__(self, table_name, columns, from_sql_string, filters,  had_changed):
        self.table_name = table_name
        self.columns = columns
        self.filters = filters
        self.from_sql_string = from_sql_string
        self.had_changed = had_changed

    def __getitem__(self, key):
        if isinstance(key, Column):
            if key.dtype != 'BOOL':
                raise Exception('Can only filter/where using column of type BOOL. got %s' % (key.dtype))
            return self.where(key)
        if isinstance(key, list):
            if all(map(lambda x: isinstance(x, str), key)) == False:
                raise Exception('List must be all strings. got %s' % (key))
            if all(map(lambda x: x in self.columns, key)) == False:
                raise Exception('All columns names must be a column in the table. got %s' % (key))
            return self.select(key)

        c = copy(self.columns[key])
        return c

    def __setitem__(self, key, newvalue):
        if not isinstance(newvalue, Column):
            raise Exception('trying to set table column with wrong type. expected type: %s, got type: %s' % (str(type(Column)), str(type(newvalue))))
        self.had_changed = True
        self.columns[key] = newvalue
    
    def __getattr__(self, attribute_name):
        return self[attribute_name]
    
    def __copy__(self):
        columns_copy = {}
        for c in self.columns.keys(): 
            columns_copy[c] = self[c]  # column deep copy will occur in __getitem__
        
        filters_copy = []
        for f in self.filters: filters_copy.append(copy(f))

        result_table = create_table(table_name=self.table_name,
                             from_sql_string=self.from_sql_string, 
                             had_changed=self.had_changed,
                             columns=columns_copy,
                             filters=filters_copy)
        return result_table

    def reset_index(self, level=None, drop=False, inplace=False, col_level=0, col_fill=''):
        return copy(self)
    
    def to_frame(self):
        return copy(self)

    def where(self, cond_column):
        self.had_changed = True
        new_table = copy(self)
        new_table.filters.append(cond_column)
        return new_table

    def select(self, columns_names):
        self.had_changed = True
        new_table = copy(self)
        # filter only selected columns from columns dictionary
        new_table.columns = \
            {col_name:col_val for (col_name, col_val) in new_table.columns.items() if col_name in columns_names}
        return new_table
    
    def merge(self, right, how='inner', on=None, left_on=None, right_on=None):
        if not isinstance(right, Table):
            raise Exception("merge expects right to be of type: %s, got: %s" %  (str(type(Table)), str(type(right))))
        if how not in ['left', 'inner']:
            raise Exception("merge 'how' value must be in [‘left’, ‘inner’]")

        left = copy(self)
        right = copy(right)
        if len(set(left.columns.keys()) & set(right.columns.keys())) > 1:
            raise Exception("merge got duplicates columns in both tables (except 'on' value)")
        
        left_on_column = None
        right_on_column = None
        if on and not left_on and not right_on:
            left_on_column = on
            right_on_column = on
        elif left_on and right_on and not on:
            left_on_column = left_on
            right_on_column = right_on
        else:
            raise Exception("got unexpected on/left_on/right_on values.")

        if not isinstance(left_on_column, str) or \
            not isinstance(right_on_column, str):
            raise Exception("'on/left_on/right_on' must be str")
        
        if left_on_column not in left.columns or right_on_column not in right.columns:
            raise Exception("merge 'on/left_on/right_on' value must be in both tables as column")
        
        left_columns = dict(zip(left.columns.keys(), map(lambda x: left[x], left.columns.keys())))
        right_columns = dict(zip(right.columns.keys(), map(lambda x: right[x], right.columns.keys())))

        # creating new table columns
        if left_on_column == right_on_column:
            right_columns.pop(on)
        new_table_columns = {**left_columns, **right_columns}

        # creating new table sql string
        single_select_field_format = 't1.%s AS %s'
        selected_fields_left = ', '.join(list(map(lambda x: single_select_field_format % (x, x), left_columns.keys())))
        
        single_select_field_format = 't2.%s AS %s'
        selected_fields_right = ', '.join(list(map(lambda x: single_select_field_format % (x, x), right_columns.keys())))
        
        selected_fields = selected_fields_left
        if selected_fields_right:
            selected_fields += ', ' + selected_fields_right
        
        new_table_sql_string = f'SELECT {selected_fields} FROM ({left.get_sql_string()}) AS t1 {how.upper()} JOIN ({right.get_sql_string()}) AS t2 ON t1.{left_on_column}=t2.{right_on_column}'
        
        return create_table(table_name='Temp',
                    columns=new_table_columns,
                    from_sql_string=new_table_sql_string)

    def groupby(self, by):
        def __get_column_key(col):
            for k in self.columns.keys():
                if self.columns[k].sql_string==col.sql_string: return k
            raise Exception('groupby got column that is not in table')
                
        groupings = None
        if isinstance(by, str):
            groupings = {by:self[by]}
        elif isinstance(by, Column):
            groupings = {__get_column_key(by): copy(by)}
        elif isinstance(by, list):
            groupings = {}
            for b in by:
                if isinstance(b, str): groupings[b] = self[b]
                elif isinstance(b, Column): groupings[__get_column_key(by)] = copy(b)
                else: raise Exception(f'groupby got unexpected type. expect str or column, got: {str(type(b))}')
        else:
            raise Exception("groupby 'by' value must be str OR list[str] OR Column OR list[Column]")
        
        return GroupedTable(copy(self), groupings=groupings)
        
    def get_sql_string(self):
        if self.from_sql_string and not self.had_changed:
            return self.from_sql_string
        
        from_field = None
        selected_fields = None
        if self.from_sql_string:
            from_field = f'({self.from_sql_string}) AS {self.table_name}'
        else:
            from_field = self.table_name

        single_select_field_format = '(%s) AS %s'
        selected_fields = ', '.join(list(map(lambda x: single_select_field_format % (self[x].sql_string, x), self.columns.keys())))

        single_where_field_format = '(%s)'
        where_cond = ' AND '.join(list(map(lambda c: single_where_field_format % (c.sql_string), self.filters)))
        
        if where_cond:
            return f'SELECT {selected_fields} FROM {from_field} WHERE {where_cond} '
        else:
            return f'SELECT {selected_fields} FROM {from_field}'
            



def create_table_from_schema(table_name, schema) -> Table:
    columns = {}
    for column_name in schema.keys():
        columns[column_name] = Column(dtype=schema[column_name], 
                                    sql_string=column_name)
    return create_table(table_name=table_name, columns=columns)

def create_table(table_name, columns={}, from_sql_string=None, filters=[], had_changed=False) -> Table:
    return Table(
        table_name=table_name,
        columns=columns,
        from_sql_string=from_sql_string,
        filters=filters,
        had_changed=had_changed)


import pandas as pd
from pandas.testing import assert_frame_equal
import pytest


def assert_dataframes_equals(expected, actual):
    assert expected.shape==actual.shape
    assert set(expected.columns) == set(actual.columns)
    columns_order = list(expected.columns)
    a = actual[columns_order].sort_values(by=list(actual.columns)).reset_index(drop=True)
    e = expected[columns_order].sort_values(by=list(actual.columns)).reset_index(drop=True)
    assert_frame_equal(e, a, check_dtype=False)

def assert_(df):
    actual_query_string = df.df_sql_convert_table.get_sql_string()
    try:
        df_actual = pd.read_sql_query(actual_query_string, pytest.sql_connection, parse_dates=['random_datetime'])
        df_expected = df.df_pandas
        assert_dataframes_equals(df_expected, df_actual)
    except Exception as e:
        # print(actual_query_string)
        raise e


from copy import copy
import operator
from pandas_to_sql.utils.pandas_dataframe_intercepter import PandasDataFrameIntercepter
from pandas_to_sql.table import create_table, Table

class PandasIntercepter:
    def __init__(self, pandas):
        self.pandas = pandas
    
    def concat(self, objs, axis=0):
        objs_pandas = list(map(lambda x: x.df_pandas, objs))
        a = self.pandas.concat(objs_pandas, axis=axis)
        objs_sql_convert = list(map(lambda x: x.df_sql_convert_table, objs))
        b = concat(objs_sql_convert, axis=axis)
        return PandasDataFrameIntercepter(a,b)


def concat(objs, axis=0):
    if axis != 0:
        raise Exception(f"supporting only axis==0")
    for df in objs:
        if not isinstance(df, Table):
            raise Exception(f'expected Table. got: {str(type(df))}')
    
    first = None
    for columns in list(map(lambda t: set(t.columns.keys()), objs)):
        if not first:
            first = columns
        else:
            if columns != first:
                raise Exception(f"expected all dataframes to have same columns")
    
    all_tables_sql_string = list(map(lambda x: x.get_sql_string(), objs))
    new_table_sql_string = ' UNION ALL '.join(all_tables_sql_string)
    return create_table(table_name='Temp',
                        columns=copy(objs[0]).columns,
                        from_sql_string=new_table_sql_string)





from copy import copy
from pandas_to_sql.column import Column

class GroupedTable:
    table = None
    groupings = None
    
    def __init__(self, table, groupings):
        self.table = table
        self.groupings = groupings

    def __copy__(self):
        return GroupedTable(copy(self.table), copy(self.groupings))

    def __getitem__(self, key):
        if isinstance(key, Column):
            raise Exception('Cant filter/where GroupedTable')
        if isinstance(key, list):
            return GroupedTable(self.table[key], copy(self.groupings))
        if isinstance(key, str):
            return GroupedTable(self.table[[key]], copy(self.groupings))
        raise Exception(f'GroupedTable __getitem__ key type not supported. type: {str(type(key))}')

    def __setitem__(self, key, newvalue):
        raise Exception(f'GroupedTable __setitem__ not supported')
        
    def __getattr__(self, attribute_name):
        return self[attribute_name]
    
    def mean(self):
        return self.agg(dict(map(lambda k: (k,'mean'),self.table.columns.keys())))

    def count(self):
        return self.agg(dict(map(lambda k: (k,'count'),self.table.columns.keys())))

    def sum(self):
        return self.agg(dict(map(lambda k: (k,'sum'),self.table.columns.keys())))

    def agg(self, v):
        if isinstance(v, str):
            return self.agg(dict(zip(self.table.columns.keys(), v)))
        elif isinstance(v, list):
            return self.agg(dict(zip(self.table.columns.keys(), v)))
        elif isinstance(v, dict):
            if len( set(v.keys()) & set(self.groupings.keys()) ) > 0:
                raise Exception("grouped table doesnt support same column in 'on' and 'select'")
            self_table_copy = copy(self.table)
            # create groupby columns query
            groupby_select_columns = {}
            for column_name in v.keys():
                column = self_table_copy[column_name]
                operations = v[column_name] if isinstance(v[column_name], list) else [v[column_name]]
                for operation in operations:
                    join_str_seperator = None
                    operation_column_name_override = None
                    dtype = None
                    
                    if callable(operation) and operation.__qualname__=='str.join':
                        join_str_seperator = operation.__self__
                        operation_column_name_override = 'join'
                        operation = 'group_concat'
#                     if not isinstance(operation, str):
#                         raise Exception(f"groupby agg support only str name for operations or ','.join. got: {type(operation)}")
#                     SUPPORTED_OPERATIONS = ['count','sum','mean','avg']
#                     if operation not in SUPPORTED_OPERATIONS:
#                         raise Exception(f"groupby operation '{operation}' is not supported. supported: {', '.join(SUPPORTED_OPERATIONS)}")

                    operation = operation.lower()
                    
                    if operation=='mean':
                        operation = 'avg'
                        operation_column_name_override = 'mean'    
                    elif operation=='sum' and column.dtype=='VARCHAR':
                        dtype = 'VARCHAR'
                        operation = 'group_concat'
                        join_str_seperator = ''
                        operation_column_name_override = 'sum'
                    elif operation=='count' or (operation=='sum' and column.dtype=='INT'):
                        dtype = 'INT'
                    else:
                        dtype = 'FLOAT'
                    
                    new_column_name = f'{column_name}_{operation_column_name_override or operation}'
                    new_sql_string = f'{operation}({column.sql_string})'
                    if operation=='group_concat':
                        new_sql_string = f"{operation}({column.sql_string},'{join_str_seperator}')"
                    groupby_select_columns[new_column_name] = Column(dtype=dtype,
                                            sql_string=new_sql_string)
            groupby_select_columns.update(self.groupings)
            
            self_table_copy.columns = groupby_select_columns

            # create new table columns
            new_table_columns = {}
            for k in groupby_select_columns.keys():
                new_table_columns[k] = Column(dtype=groupby_select_columns[k].dtype,
                                        sql_string=k)

            grouping_field = ', '.join(list(map(lambda k: self.groupings[k].sql_string, self.groupings.keys())))
            
            from pandas_to_sql.table import create_table
            return create_table(table_name='Temp',
                         columns=new_table_columns,
                         from_sql_string=f'{self_table_copy.get_sql_string()} GROUP BY {grouping_field}',
                         had_changed=False)




import pandas as pd
from pandas_to_sql.table import Table
from pandas_to_sql.column import Column
from pandas_to_sql.grouped_table import GroupedTable
from pandas_to_sql.utils.pandas_dataframe_intercepter import PandasDataFrameIntercepter
from copy import copy

## Conventions
def flatten_grouped_dataframe(df):
    if not isinstance(df, PandasDataFrameIntercepter):
        raise Exception(f"can only get type {str(type(PandasDataFrameIntercepter))}")
    
    df_c = copy(df.df_pandas)
    if isinstance(df_c, pd.core.series.Series):
        series_name = df_c.name
        new_col_name = list(filter(lambda k: k.startswith(series_name), df.df_sql_convert_table.columns.keys()))[0]
        df_c = df_c.reset_index().rename(columns={series_name: new_col_name})
    else:
        df_c.columns = df_c.columns.map('_'.join)
        df_c = df_c.reset_index()
    return PandasDataFrameIntercepter(df_c, copy(df.df_sql_convert_table))


## Types
def convert_df_type(col_type):
    if pd.api.types.is_bool_dtype(col_type): return 'BOOL'
    elif pd.api.types.is_integer_dtype(col_type): return 'INT'
    elif pd.api.types.is_numeric_dtype(col_type): return 'FLOAT'
    elif pd.api.types.is_string_dtype(col_type): return 'VARCHAR'
    elif pd.api.types.is_datetime64_any_dtype(col_type): return 'TIMESTAMP'
    else: raise Exception(f"could not convert column type. got: {str(col_type)}")

def create_schema_from_df(df):        
    schema = {}
    for col_name, col_type in df.dtypes.items():
        schema[col_name] = convert_df_type(col_type)
    return schema


from copy import copy
import operator

class PandasDataFrameIntercepter:
    def __init__(self, df_pandas, df_sql_convert_table):
        self.df_pandas = df_pandas
        self.df_sql_convert_table = df_sql_convert_table

    def __repr__(self):
        return self.df_pandas.__repr__()

    def __format__(self):
        return self.df_pandas.__format__()

    def __str__(self):
        return self.df_pandas.__str__()
    
    @staticmethod
    def get_attr_for_df_pandas_if_needed(obj):
        if isinstance(obj, PandasDataFrameIntercepter):
            return object.__getattribute__(obj, 'df_pandas')
        else:
            return obj

    @staticmethod
    def get_attr_for_df_sql_convert_table_if_needed(obj):
        if isinstance(obj, PandasDataFrameIntercepter):
            return object.__getattribute__(obj, 'df_sql_convert_table')
        else:
            return obj

    def __getattribute__(self, name):
        if name in ['df_pandas', 'df_sql_convert_table']:
            return object.__getattribute__(self, name)
        
        df_sql_convert_table_attr =  object.__getattribute__(self.df_sql_convert_table, name)
        if name=='get_sql_string' and hasattr(df_sql_convert_table_attr, '__call__'):
            return lambda *args, **kwargs: df_sql_convert_table_attr(*args, **kwargs)

        df_pandas_attr =  object.__getattribute__(self.df_pandas, name)
        if name=='columns' and not hasattr(df_pandas_attr, '__call__'):
            return df_pandas_attr
              
        if hasattr(df_sql_convert_table_attr, '__call__'):
            def _(*args, **kwargs):
                def __dictionary_map_values(d, func):
                    return {k: func(v) for k, v in d.items()}
                
                args_df_pandas = tuple(map(PandasDataFrameIntercepter.get_attr_for_df_pandas_if_needed, args))
                args_obj_new = tuple(map(PandasDataFrameIntercepter.get_attr_for_df_sql_convert_table_if_needed, args))
                
                kwargs_df_pandas = __dictionary_map_values(kwargs, PandasDataFrameIntercepter.get_attr_for_df_pandas_if_needed)
                kwargs_obj_new = __dictionary_map_values(kwargs, PandasDataFrameIntercepter.get_attr_for_df_sql_convert_table_if_needed)
                
                a = df_pandas_attr(*args_df_pandas, **kwargs_df_pandas)
                b = df_sql_convert_table_attr(*args_obj_new, **kwargs_obj_new)
                return PandasDataFrameIntercepter(a, b)
            return _
        else:           
            return PandasDataFrameIntercepter(df_pandas_attr, df_sql_convert_table_attr)
    
    def __getitem__(self, key):
        a = self.df_pandas[PandasDataFrameIntercepter.get_attr_for_df_pandas_if_needed(key)]
        b = self.df_sql_convert_table[PandasDataFrameIntercepter.get_attr_for_df_sql_convert_table_if_needed(key)]
        return PandasDataFrameIntercepter(a, b)
    
    def __setitem__(self, key, newvalue):
        self.df_pandas[key] = PandasDataFrameIntercepter.get_attr_for_df_pandas_if_needed(newvalue)
        self.df_sql_convert_table[key] = PandasDataFrameIntercepter.get_attr_for_df_sql_convert_table_if_needed(newvalue)
        return PandasDataFrameIntercepter(self.df_pandas, self.df_sql_convert_table)

    def __getattr__(self, attribute_name):
        a = self.df_pandas[attribute_name]
        b = self.df_sql_convert_table[attribute_name]
        return PandasDataFrameIntercepter(a, b)
    
    def __copy__(self):
        return PandasDataFrameIntercepter(copy(self.df_pandas), copy(self.df_sql_convert_table))
    
    @staticmethod
    def run_operation_and_return(left, right, op):
        left_ = PandasDataFrameIntercepter.get_attr_for_df_pandas_if_needed(left)
        right_ = PandasDataFrameIntercepter.get_attr_for_df_pandas_if_needed(right)
        a = op(left_, right_)

        left_ = PandasDataFrameIntercepter.get_attr_for_df_sql_convert_table_if_needed(left)
        right_ = PandasDataFrameIntercepter.get_attr_for_df_sql_convert_table_if_needed(right)
        b = op(left_, right_)
        return PandasDataFrameIntercepter(a, b)

    @staticmethod
    def run_operation_single_and_return(obj, op):
        a = PandasDataFrameIntercepter.get_attr_for_df_pandas_if_needed(obj)
        b = PandasDataFrameIntercepter.get_attr_for_df_sql_convert_table_if_needed(obj)
        a = op(a)
        b = op(b)
        return PandasDataFrameIntercepter(a, b)
    
    # comparisons
    def __lt__(self,other):
        return PandasDataFrameIntercepter.run_operation_and_return(self, other, operator.lt)

    def __le__(self,other):
        return PandasDataFrameIntercepter.run_operation_and_return(self, other, operator.le)

    def __gt__(self,other):
        return PandasDataFrameIntercepter.run_operation_and_return(self, other, operator.gt)

    def __ge__(self,other):
        return PandasDataFrameIntercepter.run_operation_and_return(self, other, operator.ge)

    def __eq__(self,other):
        return PandasDataFrameIntercepter.run_operation_and_return(self, other, operator.eq)

    def __ne__(self,other):
        return PandasDataFrameIntercepter.run_operation_and_return(self, other, operator.ne)

    def __abs__(self):
        return PandasDataFrameIntercepter.run_operation_single_and_return(self, operator.abs)
        
    def __neg__(self):
        return PandasDataFrameIntercepter.run_operation_single_and_return(self, operator.neg)

    def __invert__(self):
        return PandasDataFrameIntercepter.run_operation_single_and_return(self, operator.invert)

    def __contains__(self, r):
        return PandasDataFrameIntercepter.run_operation_and_return(self, r, operator.contains)

    # numeric 
    def __add__(self, r):
        return PandasDataFrameIntercepter.run_operation_and_return(self, r, operator.add)
    
    def __sub__(self, r):
        return PandasDataFrameIntercepter.run_operation_and_return(self, r, operator.sub)
    
    def __mul__(self, r):
        return PandasDataFrameIntercepter.run_operation_and_return(self, r, operator.mul)
    
    # def __matmul__(self, r):
    #     return PandasDataFrameIntercepter.run_operation_and_return(self, r, operator.matmul)

    def __truediv__(self, r):
        return PandasDataFrameIntercepter.run_operation_and_return(self, r, operator.truediv)
    
    def __floordiv__(self, r):
        return PandasDataFrameIntercepter.run_operation_and_return(self, r, operator.floordiv)
    
    def __mod__(self, r):
        return PandasDataFrameIntercepter.run_operation_and_return(self, r, operator.mod)
    
    def __pow__(self, r):
        return PandasDataFrameIntercepter.run_operation_and_return(self, r, operator.pow)
    
    def __and__(self, r):
        return PandasDataFrameIntercepter.run_operation_and_return(self, r, operator.and_)
    
    def __or__(self, r):
        return PandasDataFrameIntercepter.run_operation_and_return(self, r, operator.or_)

    # numeric r
    def __radd__(self, l):
        return PandasDataFrameIntercepter.run_operation_and_return(l, self, operator.add)
    
    def __rsub__(self, l):
        return PandasDataFrameIntercepter.run_operation_and_return(l, self, operator.sub)
    
    def __rmul__(self, l):
        return PandasDataFrameIntercepter.run_operation_and_return(l, self, operator.mul)
    
    def __rmatmul__(self, l):
        return PandasDataFrameIntercepter.run_operation_and_return(l, self, operator.matmul)

    def __rtruediv__(self, l):
        return PandasDataFrameIntercepter.run_operation_and_return(l, self, operator.truediv)
    
    def __rfloordiv__(self, l):
        return PandasDataFrameIntercepter.run_operation_and_return(l, self, operator.floordiv)
    
    def __rmod__(self, l):
        return PandasDataFrameIntercepter.run_operation_and_return(l, self, operator.mod)
    
    def __rpow__(self, l):
        return PandasDataFrameIntercepter.run_operation_and_return(l, self, operator.pow)
    
    def __rand__(self, l):
        return PandasDataFrameIntercepter.run_operation_and_return(l, self, operator.and_)
    
    def __ror__(self, l):
        return PandasDataFrameIntercepter.run_operation_and_return(l, self, operator.or_)
  
    
    
    



import os
import setuptools

v = os.environ['RELEASE_VERSION']
print('Version: ', v)

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pandas-to-sql", # Replace with your own username
    version=v,
    author="Amir",
    author_email="amirpupko@gmail.com",
    description="Convert pandas dataframe manipulations to sql query string",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AmirPupko/pandas-to-sql",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


import pandas as pd
import numpy as np


def random_datetimes_or_dates(start, end, out_format='datetime', n=10):
    '''   
    unix timestamp is in ns by default. 
    I divide the unix time value by 10**9 to make it seconds (or 24*60*60*10**9 to make it days).
    The corresponding unit variable is passed to the pd.to_datetime function. 
    Values for the (divide_by, unit) pair to select is defined by the out_format parameter.
    for 1 -> out_format='datetime'
    for 2 -> out_format=anything else
    '''
    (divide_by, unit) = (
        10**9, 's') if out_format == 'datetime' else (24*60*60*10**9, 'D')

    start_u = start.value//divide_by
    end_u = end.value//divide_by

    return pd.to_datetime(np.random.randint(start_u, end_u, n), unit=unit)


def random_timedelta(start, end, n, unit='D', seed=None):
    if not seed:  # from piR's answer
        np.random.seed(0)

    ndays = (end - start).days + 1
    return pd.to_timedelta(np.random.rand(n) * ndays, unit=unit)


def create_fake_dataset(start=pd.to_datetime('2015-01-01'), end=pd.to_datetime('2018-01-01')):
    df = pd.DataFrame()
    df_size = 1000
    df_random_columns = {
        'random_int': 'INT',
        'random_float': 'FLOAT',
        'random_bool': 'BOOL',
        'random_datetime': 'TIMESTAMP',
        'random_str': 'VARCHAR',
    }
    df['random_int'] = np.random.randint(1, 6, df_size)
    df['random_float'] = np.random.randn(df_size)
    df['random_bool'] = np.random.randn(df_size) > 0
    df['random_datetime'] = random_datetimes_or_dates(start, end, n=df_size)
    df['random_str'] = pd.util.testing.rands_array(10, df_size)
    return df, df_random_columns
